#include <windows.h>
#include <d3d9.h>
#include <D3dx9tex.h>
#include <D3dx9math.h>
#include <cstddef>
#include <stdio.h>
#include <dxerr.h>

IDirect3D9					*g_direct3D = NULL;
IDirect3DDevice9			*g_device = NULL;
IDirect3DTexture9			*g_tex0 = NULL;
IDirect3DTexture9			*g_tex1 = NULL;
IDirect3DTexture9			*g_tex2 = NULL;
IDirect3DVertexDeclaration9 *g_vertexDecl = NULL;
IDirect3DSurface9			*g_defaultRT = NULL;
IDirect3DTexture9			*g_rtt = NULL;
IDirect3DSurface9			*g_rttSurf = NULL;

struct Vertex
{
	D3DXVECTOR3 Position;
	D3DCOLOR Color;
	D3DXVECTOR2 Texture;
};

LRESULT WINAPI WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	switch(msg)
	{
	case WM_DESTROY:
		PostQuitMessage(0);
		return(0);
	}

	return(DefWindowProc(hwnd, msg, wParam, lParam));
}

void DrawTexturedQuad(IDirect3DTexture9* tex, int x, int y, int w, int h)
{
	const float offset = -0.5f;

	Vertex verts[6] = {
		{ D3DXVECTOR3(x+offset,y+offset,0), D3DCOLOR_ARGB(255,255,255,255), D3DXVECTOR2(0,0) },
		{ D3DXVECTOR3(x+w+offset,y+offset,0), D3DCOLOR_ARGB(255,255,255,255), D3DXVECTOR2(1,0) },
		{ D3DXVECTOR3(x+offset,y+h+offset+offset,0), D3DCOLOR_ARGB(255,255,255,255), D3DXVECTOR2(0,1) },
		{ D3DXVECTOR3(x+w+offset,y+h+offset,0), D3DCOLOR_ARGB(255,255,255,255), D3DXVECTOR2(1,1) },
	};

	g_device->SetTexture(0, tex);
	g_device->DrawPrimitiveUP(D3DPT_TRIANGLESTRIP, 2, &verts[0], sizeof(Vertex));
}

void Draw()
{
	g_device->BeginScene();

	// Common states we always use
	g_device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);
	g_device->SetRenderState(D3DRS_LIGHTING, FALSE);
	g_device->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);

	// Render to 512x512 texture
	g_device->SetRenderTarget(0, g_rttSurf);
	D3DXMATRIX proj512x512;
	D3DXMatrixOrthoOffCenterRH(&proj512x512, 0, 512, 512, 0, -10, 10);
	g_device->SetTransform(D3DTS_PROJECTION, &proj512x512);

	// Clear all components to zero
	g_device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_ARGB(0,0,0,0), 0, 0);

	// Premultiply alpha in the source texture using texture stage
	g_device->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	g_device->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	g_device->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_TEXTURE|D3DTA_ALPHAREPLICATE);

	// Set blend mode to (ONE, INVSRCALPHA)
	g_device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
	g_device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

	// Draw textures 1 & 2
	DrawTexturedQuad(g_tex1, 0, 0, 512, 512);
	DrawTexturedQuad(g_tex2, 0, 0, 512, 512);

	// Now switch to the main render target (1024x512)
	g_device->SetRenderTarget(0, g_defaultRT);
	D3DXMATRIX proj1024x512;
	D3DXMatrixOrthoOffCenterRH(&proj1024x512, 0, 1024, 512, 0, -10, 10);

	// Clear to some random color
	g_device->Clear(0, NULL, D3DCLEAR_TARGET, D3DCOLOR_XRGB(0, 0, 255), 0, 0);

	// Reset texture stage settings to default
	g_device->SetTextureStageState(0, D3DTSS_COLOROP, D3DTOP_MODULATE);
	g_device->SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
	g_device->SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_CURRENT);
	g_device->SetTransform(D3DTS_PROJECTION, &proj1024x512);	

	// Left side. (render to texture side)
	// Draw the back texture, 0.png -- use normal alpha blending
	// (doesn't matter since it's opaque)
	g_device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
	g_device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

	DrawTexturedQuad(g_tex0, 0, 0, 512, 512);

	// Draw the offscreen texture we have generated.
	// It uses premultiplied alpha so we have to blend appropriately
	g_device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_ONE);
	g_device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

	DrawTexturedQuad(g_rtt, 0, 0, 512, 512);

	// Right side. (control side)
	// Now go back and draw each texture on top of one another w/ default alpha blending.
	g_device->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
	g_device->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

	DrawTexturedQuad(g_tex0, 512, 0, 512, 512);
	DrawTexturedQuad(g_tex1, 512, 0, 512, 512);
	DrawTexturedQuad(g_tex2, 512, 0, 512, 512);

	// Done with scene, swap buffers, etc.
	g_device->EndScene();
	g_device->Present(NULL, NULL, NULL, NULL);
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nShow)
{
	MSG msg;

	WNDCLASSEX wc = {sizeof(WNDCLASSEX), CS_VREDRAW|CS_HREDRAW|CS_OWNDC, 
		WndProc, 0, 0, hInstance, NULL, NULL, (HBRUSH)(COLOR_WINDOW+1), 
		NULL, "DX9_TUTORIAL1_CLASS", NULL};

	RegisterClassEx(&wc);

	RECT wr = { 0, 0, 1024, 512 };
	AdjustWindowRect(&wr, WS_OVERLAPPEDWINDOW, false);

	HWND hMainWnd = CreateWindow("DX9_TUTORIAL1_CLASS",
		"DirectX 9 Bare Bones Tutorial 1", 
		WS_OVERLAPPEDWINDOW, 100, 100, wr.right-wr.left, wr.bottom-wr.top,
		NULL, NULL, hInstance, NULL);

	g_direct3D = Direct3DCreate9(D3D_SDK_VERSION);

	D3DPRESENT_PARAMETERS PresentParams;
	memset(&PresentParams, 0, sizeof(D3DPRESENT_PARAMETERS));

	PresentParams.Windowed = TRUE;
	PresentParams.SwapEffect = D3DSWAPEFFECT_DISCARD;

	g_direct3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hMainWnd,
		D3DCREATE_SOFTWARE_VERTEXPROCESSING, &PresentParams,
		&g_device);

	D3DVERTEXELEMENT9 elements[] = {
		{ 0, offsetof(Vertex, Position), D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
		{ 0, offsetof(Vertex, Color), D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0 },
		{ 0, offsetof(Vertex, Texture), D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },
		D3DDECL_END()
	};

	g_device->CreateVertexDeclaration(elements, &g_vertexDecl);
	g_device->SetVertexDeclaration(g_vertexDecl);

	D3DXCreateTextureFromFile(g_device, "0.png", &g_tex0);
	D3DXCreateTextureFromFile(g_device, "1.png", &g_tex1);
	D3DXCreateTextureFromFile(g_device, "2.png", &g_tex2);

	// Create render target texture
	g_device->CreateTexture(512, 512, 0, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &g_rtt, NULL);
	g_rtt->GetSurfaceLevel(0, &g_rttSurf);
	g_device->GetRenderTarget(0, &g_defaultRT);

	ShowWindow(hMainWnd, nShow);
	UpdateWindow(hMainWnd);

	bool done=false;
	while(!done)
	{
		while( PeekMessage( &msg, NULL, 0, 0, PM_REMOVE ))
		{
			if(msg.message == WM_QUIT)
			{
				done=true;
				break;
			}
			TranslateMessage( &msg );
			DispatchMessage( &msg );
		}

		Draw();
	}

	g_device->Release();
	g_direct3D->Release();

	return(0);
}

